# jupyter-require
# Copyright 2019 Marek Cermak <macermak@redhat.com>
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for managing linked JavaScript scripts and CSS styles."""

import logging
import string

import daiquiri

from datetime import datetime

from collections import OrderedDict
from pathlib import Path

from typing import List, Union

from IPython import get_ipython
from IPython.core.display import display, Javascript

from ipykernel.comm import Comm

logger = daiquiri.getLogger()


Jupyter = get_ipython()
"""Current InteractiveShell instance."""

_HERE = Path(__file__).parent


class RequireJS(object):

    __LIBS = OrderedDict()
    """Required libraries."""
    __SHIM = OrderedDict()
    """Shim for required libraries."""

    def __init__(self, required: dict = None, shim: dict = None):
        """Initialize RequireJS."""
        # comm messages
        self._msg = None
        self._msg_received = None

        # comms
        self._config_comm = None
        self._execution_comm = None
        self._safe_execution_comm = None

        # update with default required libraries
        self.__LIBS.update(required or {})
        self.__SHIM.update(shim or {})

        # check if running in Jupyter notebook
        self._is_notebook = Jupyter and Jupyter.has_trait('kernel')

    def __call__(self, library: str, path: str, *args, **kwargs):
        """Links JavaScript library to Jupyter Notebook.

        The library is linked using requireJS such as:

        ```javascript
        require.config({ paths: {<key>: <path>} });
        ```

        Please note that <path> does __NOT__ contain `.js` suffix.

        :param library: str, key to the library
        :param path: str, path (url) to the library without .js suffix
        """
        self.config({library: path}, shim=kwargs.pop('shim', {}))

    @property
    def libs(self) -> dict:
        """Get custom loaded libraries."""
        return dict(self.__LIBS)

    @property
    def shim(self) -> dict:
        """Get shim defined in requireJS config."""
        return dict(self.__SHIM)

    def display_context(self):
        """Print defined libraries."""
        _ = self  # ignore

        return display(Javascript("""
            const context = require.s.contexts._.defined;

            $(element).html(
                JSON.stringify(Object.keys(context).sort())
                .replace(/,/g, '<br>'));
        """))

    def config(self, paths: dict, shim: dict = None):
        """Links JavaScript libraries to Jupyter Notebook.

        This is Python binding for JS RequireJS `require.config` call.

        For convenience, the paths and shim are split into two separate arguments. If you
        with to use the function in a more standard JS way, consider the following example.

        Example:
        ```
        require.config(**{
            "paths": {
                <key>: <path>
            },
            "shim": {
                <key>: [<dependencies>]
            }
        });
        ```

        Please note that <path> does __NOT__ contain `.js` suffix.
        """
        self.__LIBS.update(paths)
        self.__SHIM.update(shim or {})

        # data to be passed to require.config()
        self._msg = {'paths': self.__LIBS, 'shim': self.__SHIM}

        self._config_comm.send(data=self._msg)

    def pop(self, lib: str):
        """Remove JavaScript library from requirements.

        :param lib: key as passed to `config()`
        """
        self.__LIBS.pop(lib)
        self.__SHIM.pop(lib)

    @classmethod
    def reload(cls, clear=False):
        """Reload and create new require object."""
        global require

        if clear:
            cls.__LIBS.clear()
            cls.__SHIM.clear()

        libs = require.libs if not clear else []
        shim = require.shim if not clear else []

        require = cls(required=libs, shim=shim)

        if require._is_notebook:
            require._initialize_comms()

        require.__doc__ = RequireJS.__call__.__doc__

    def _initialize_comms(self):
        """Initialize Python-JavaScript comms."""
        now = datetime.now()

        self._config_comm = create_comm(
            target='config',
            comm_id=f'config.JupyterRequire#{datetime.timestamp(now)}',
            callback=self._store_callback)

        self._execution_comm = create_comm(
            target='execute',
            comm_id=f'execute.JupyterRequire#{datetime.timestamp(now)}',
            callback=self._store_callback)

        self._safe_execution_comm = create_comm(
            target='safe_execute',
            comm_id=f'safe_execute.JupyterRequire#{datetime.timestamp(now)}',
            callback=self._store_callback)

        # initial configuration
        self.config(paths={})

    def _store_callback(self, msg):
        """Store callback from comm."""
        self._msg_received = msg


class JSTemplate(string.Template):
    """Custom JS string template."""

    delimiter = "$$"

    def __init__(self, template: str):
        super().__init__(template)

        self._safe_substitute = self.safe_substitute
        self._substitute = self.substitute

        # prototype
        def safe_substitute(*args, **kws):
            """Safely substitute JS template variables."""
            kwargs = {
                key: sub if sub is not None else 'null'
                for key, sub in kws.items()
            }

            return self._safe_substitute(*args, **kwargs)

        # prototype
        def substitute(*args, **kws):
            """Substitute JS template variables."""
            kwargs = {
                key: sub if sub is not None else 'null'
                for key, sub in kws.items()
            }

            return self._substitute(*args, **kwargs)

        self.safe_substitute = safe_substitute
        self.substitute = substitute


def create_comm(target: str,
                data: dict = None,
                callback: callable = None,
                **kwargs):
    """Create ipykernel message comm."""
    # create comm on python site
    comm = Comm(target_name=target, data=data, **kwargs)
    comm.on_msg(callback)

    return comm


def execute_with_requirements(script: str, required: Union[list, dict], configured=True, **kwargs):
    """Link required libraries and execute JS script.

    :param script: JS script to be executed
    :param required: list or dict (for requireJS config) of requirements
    :param configured: bool, whether requirements are already configured

        This speeds up the execution, so if the requirements are already configured,
        do not run configuration again.

        Assume True, as user is expected to run `require.config()`
        at the initialization time.

    :param kwargs: optional keyword arguments for template substitution
    """
    if not configured:
        if isinstance(required, dict):
            require.config(required, **kwargs)
        else:
            raise TypeError(
                f"Attribute `required` expected to be dict, got {type(required)}.")

    required: list = required if isinstance(required, list) else list(required.keys())

    params = kwargs.pop('params', []) or required
    params = list(map(lambda s: s.rsplit('/')[-1], params))

    script = JSTemplate(script).safe_substitute(**kwargs)

    data = {
        'script': script,
        'require': required,
        'parameters': params,
    }

    if require._safe_execution_comm is None:
        raise TypeError("Comm 'execute' is not open.")

    # noinspection PyProtectedAccess
    return require._execution_comm.send(data)  # pylint: disable=protected-access


def execute(script: str, **kwargs):
    """Execute JS script.

    This function implicitly loads libraries defined in requireJS config.
    """
    required = []
    try:
        required = list(require.libs.keys())
    except NameError:  # require has not been defined yet, allowed
        pass

    return execute_with_requirements(script, required=required, **kwargs)


def safe_execute(script: str, **kwargs):
    """Execute JS script and treat it as safe script.

    Safe scripts are executed on cell creation
    and are therefore not allowed to have any requirements.
    Scripts executed with this method also persist through notebook
    reloads and are automatically loaded on app initialization.

    This function is convenient for automatic loading and linking
    of custom CSS and JS files.
    """
    script = "{ " + script + " }"  # provide local scope
    script = JSTemplate(script).safe_substitute(**kwargs)

    if require._safe_execution_comm is None:
        raise TypeError("Comm 'safe_execute' is not open.")

    # noinspection PyProtectedAccess
    return require._safe_execution_comm.send(data={'script': script})  # pylint: disable=protected-access


require = RequireJS()
require.__doc__ = RequireJS.__call__.__doc__


def _handle_comms_registered(*args, **kwargs):
    """Handle comms_registered.JupyterRequire event."""
    logger.debug("Comms registered.")


def _handle_extension_loaded(*args, **kwargs):
    """Handle extension_loaded.JupyterRequire event."""
    logger.debug("Extension loaded.")

    # Jupyter discards promises on window reload,
    # so comms have to be re-initialized
    require._initialize_comms()

    msg = "Comms initialized."
    logger.debug(msg)

    return msg


_event_handle_map = {
    'comms_registered': _handle_comms_registered,
    'extension_loaded': _handle_extension_loaded
}


def communicate(comm, open_msg):
    """Handle messages from Jupyter frontend."""
    _ = open_msg  # ignored

    logger.debug("Comm 'communicate' opened.")

    @comm.on_msg
    def handle_msg(msg):
        """Handle message."""
        logger.debug(
            "Message received: %r", msg)

        data = msg['content']['data']

        event = data.get('event', None)
        event_data = data.get('event_data', {})

        logger.debug(
            "Requested message handler.\n\tData: %r\n\tEvent: %r", data, event)

        response = {'resolved': True, 'value': None, 'success': False}
        try:
            event_type, namespace = event['type'], event['namespace']

            if namespace == 'JupyterRequire':
                handle = _event_handle_map[event_type]

                response['value'] = handle(event_data)
                response['success'] = True

            logger.debug("Success.")

        except Exception as err:
            logger.error(
                "Error: '%s': %s", event, err)
            response['value'] = err

        finally:
            require._store_callback(msg)

        comm.send(response)



# jupyter-require
# Copyright 2019 Marek Cermak <macermak@redhat.com>
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Jupyter library and magic extension for managing linked JavaScript and CSS scripts and styles."""

import logging
import time

import daiquiri
import daiquiri.formatter

from .__about__ import __version__

from .notebook import link_css
from .notebook import link_js
from .notebook import load_js
from .notebook import load_css

from .core import communicate

from .core import execute_with_requirements
from .core import execute
from .core import safe_execute

from .core import require

from IPython import get_ipython


daiquiri.setup(
    level=logging.DEBUG,
    outputs=[
        daiquiri.output.File(
            level=logging.DEBUG,
            filename='.log',
            formatter=daiquiri.formatter.ColorFormatter(
                fmt="%(asctime)s [%(process)d] %(color)s%(levelname)-8.8s %(name)s:"
                    "%(lineno)d: [JupyterRequire] %(message)s%(color_stop)s"
            )),
        daiquiri.output.Stream(
            level=logging.WARN,
            formatter=daiquiri.formatter.ColorFormatter(
                fmt="%(asctime)s [%(process)d] %(color)s%(levelname)-8.8s %(name)s:"
                    "%(lineno)d: [JupyterRequire] %(message)s%(color_stop)s"
            )
        ),
    ],
)

logger = daiquiri.getLogger()


def load_ipython_extension(ipython):
    """Load the IPython Jupyter Require extension."""
    from .magic import RequireJSMagic

    logger.debug("Loading Jupyter Require extension.")

    if not hasattr(ipython, 'kernel'):
        logger.debug("No kernel found.")
        return

    register_comm_targets(ipython.kernel)
    time.sleep(0.5)  # let the JS register the targets

    # fontawesome fas icon
    link_css(
        "https://use.fontawesome.com/releases/v5.8.1/css/all.css",
        attrs={
            'integrity': "sha384-50oBUHEmvpQ+1lW4y57PTFmhCaXp0ML5d60M1M7uH2+nqUivzIebhndOJK28anvf",
            'crossorigin': "anonymous"
        })

    # magic: %require
    ipython.register_magics(RequireJSMagic)


def register_comm_targets(kernel=None):
    """Register comm targets."""
    if kernel is None:
        kernel = get_ipython().kernel

    logger.debug("Initializing comms.")
    # noinspection PyProtectedMember
    require._initialize_comms()  # pylint: disable=protected-access

    logger.debug("Registering comm targets.")
    kernel.comm_manager.register_target('communicate', communicate)


def _handle_ipython():
    """Register with the comm target at import."""
    ipython = get_ipython()
    if ipython is None:
        return

    load_ipython_extension(ipython)


def _jupyter_nbextension_paths():
    return [{
        'section': 'notebook',
        'src': 'static',  # FIXME when migrated to node.js
        'dest': 'jupyter-require',
        'require': 'jupyter-require/extension'
    }]


_handle_ipython()



# -*- coding: utf-8 -*-

from __future__ import print_function

import os

from pathlib import Path
from setupbase import (
    log,
    create_cmdclass,
    install_npm,
    combine_commands,
    ensure_targets,
)

from setuptools import setup
from setuptools import find_packages


HERE = Path(__file__).parent
NAME = 'jupyter_require'

ABOUT = dict()
exec(Path(HERE, NAME, '__about__.py').read_text(), ABOUT)

README: str = Path(HERE, "README.rst").read_text(encoding='utf-8')
REQUIREMENTS: list = Path(HERE, 'requirements.txt').read_text().splitlines()

log.set_verbosity(log.DEBUG)
log.info('setup.py entered')
log.info('$PATH=%s' % os.environ['PATH'])

cmdclass = create_cmdclass(
    'js',
    data_files_spec=[
        ('share/jupyter/nbextensions/jupyter-require',
         NAME + '/static',
         '*.js'),
        ('share/jupyter/nbextensions/jupyter-require',
         NAME + '/static',
         '*.js.map'),
        ('etc/jupyter/nbconfig',
         'jupyter-config',
         '**/*.json'),
    ],
)
cmdclass['js'] = combine_commands(
    # FIXME when migrated to node.js
    # install_npm(
    #     path=Path(HERE, 'js'),
    #     build_dir=Path(HERE, NAME, 'static'),
    #     source_dir=Path(HERE, 'js'),
    #     build_cmd='build:all'
    # ),
    ensure_targets([
        NAME + '/static/core.js',  # FIXME when migrated to nodes.js
        NAME + '/static/display.js',  # FIXME when migrated to nodes.js
        NAME + '/static/extension.js',
        # NAME + '/static/index.js',  # FIXME when migrated to nodes.js
    ]),
)


setup(
    name=ABOUT['__title__'],
    version=ABOUT['__version__'],

    author=ABOUT['__author__'],
    author_email=ABOUT['__email__'],
    url=ABOUT['__uri__'],

    license=ABOUT['__license__'],

    description=ABOUT['__summary__'],
    long_description=README,
    long_description_content_type='text/x-rst',

    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Framework :: IPython",
        "Framework :: Jupyter",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: JavaScript",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Visualization",
        "Topic :: Utilities"
    ],

    install_requires=REQUIREMENTS,

    cmdclass=cmdclass,

    packages=find_packages(),
    include_package_data=True,

    zip_safe=False
)



# jupyter-require
# Copyright 2019 Marek Cermak <macermak@redhat.com>
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Jupyter magic for managing linked JavaScript scripts and CSS styles."""

import re

from IPython.core.magic import cell_magic
from IPython.core.magic import line_magic
from IPython.core.magic import line_cell_magic
from IPython.core.magic import magics_class
from IPython.core.magic import Magics
from IPython.core.magic import needs_local_scope

from jupyter_nbutils.utils import sanitize_namespace

from .core import execute_with_requirements
from .core import require
from .core import safe_execute

from .core import JSTemplate

from .notebook import link_css as _link_css
from .notebook import load_css as _load_css
from .notebook import link_js as _link_js
from .notebook import load_js as _load_js


def activate_js_syntax_highlight(regex: str = 'requirejs'):
    """Activates syntax highlighting for the `%%require` cells."""
    script = """
    const regex = "^%%$$regex"

    let modes = codecell.CodeCell.options_default.highlight_modes

    const mimes = ['magic_text/javascript', 'magic_javascript']

    mimes.forEach( (m) => {
        let mode = modes[m] || {}

        if ( _.isUndefined(mode.reg) )
            mode.reg = [regex]
        else
            mode.reg.push(regex)

        modes[m] = mode
    });

    Jupyter.notebook.events.one('kernel_ready.Kernel', function() {
      Jupyter.notebook.get_cells().map( (cell) => {
          if (cell.cell_type == 'code') cell.auto_highlight()
      })
    });

    console.debug(`JavaScript syntax highlight activated for '${regex}'.`)
    """

    return execute_with_requirements(script, required=['notebook/js/codecell'], regex=regex)


@magics_class
class RequireJSMagic(Magics):
    """Ipython magic for RequireJS class.

    Links JavaScript libraries to Jupyter Notebook.
    """

    def __init__(self, *args, **kwargs):
        super(RequireJSMagic, self).__init__(*args, **kwargs)

        activate_js_syntax_highlight()

    @needs_local_scope
    @line_cell_magic
    def requirejs(self, line: str, cell: str = None, local_ns=None):
        """Execute current JS cell with requirements or link required JS library.

        Line magic: Link required JS library.

        :param line: string in form '<key> <path>'
        :param local_ns: current cell namespace [optional]

        Cell magic: Execute current JS cell with requirements 

        The required libraries specified in parameters have to be defined and loaded in advance,
        `require` line magic can be used for that purpose.

        :param line: str, requirements separated by spaces
        :param cell: str, script to be executed
        :param local_ns: current cell namespace [optional]
        """
        user_ns = self.shell.user_ns
        user_ns.update(local_ns or dict())

        if cell is None:
            if not line:
                return require.display_context()

            lib, path = line \
                .strip() \
                .split(sep=' ')
            
            if not path:
                raise ValueError("Path to the library was not defined correctly.")

            return require(lib, path)

        ns = sanitize_namespace(user_ns, options={'warnings': False})

        # do not use safe substitution here
        script = JSTemplate(cell).substitute(**ns)

        required = line \
            .strip() \
            .split(sep=' ')

        return execute_with_requirements(script, required)

    @cell_magic
    def define(self, line: str, cell: str):
        """Define new module from the current cell content.
        
        :param line: module name
        :param cell: script to be defined as module by the module name
        """
        if not line:
            raise ValueError("Module name required but not provided.")

        script = """
        const module = '$$module';

        // overwrite any previously defined modules with the same name
        requirejs.undef(module);

        define(module, function(require) {
            $$cell
        });
        """

        return safe_execute(script, cell=cell, module=line)

    @line_magic
    def undef(self, line: str):
        """Undefine required libraries.

        :param line: str, libs to undefine separated by spaces
        """
        libs = line \
            .strip() \
            .split(sep=' ')

        script = """
            const libs = $$to_undefine;

            libs.forEach((lib) => {
                requirejs.undef(lib);
            });
        """

        for lib in libs:
            try:
                require.pop(lib)
            except KeyError:
                pass

        return safe_execute(script, to_undefine=libs)

    @line_magic
    def reloadjs(self, line: str):
        """Reload JS libraries.

        This is especially useful when needed to reload
        script after making changes to it.

        :param line: str, libs to reload separated by spaces
        """
        libs = line \
            .strip() \
            .split(sep=' ')

        script = """
            const libs = $$to_reload;

            libs.forEach((lib) => {
                requirejs.undef(lib);
            });
            requirejs(libs, () => {
                console.debug(`Libraries ${libs} reloaded.`);
            });
        """

        return safe_execute(script, to_reload=libs)

    @line_magic
    def link_css(self, line: str):
        """Link CSS stylesheet."""
        return _link_css(line)

    @line_magic
    def link_js(self, line: str):
        """Link JavaScript library."""
        return _link_js(line)

    @cell_magic
    def load_css(self, line: str, cell: str):
        """Create new style element and add it to the page."""
        attributes: dict = self._parse_attributes(line)

        return _load_css(cell, attributes)

    @cell_magic
    def load_js(self, line: str, cell: str):
        """Create new script element and add it to the page."""
        attributes: dict = self._parse_attributes(line)

        return _load_js(cell, attributes)

    @staticmethod
    def _parse_attributes(line: str) -> dict:
        """Parse string to return element attributes."""
        attr_def = map(
            lambda s: re.sub(r"[\"'](.*)[\"']", r"\1", s),
            line.split()
        )

        return {
            attr: val
            for attr, val in map(lambda s: s.split('='), attr_def)
        }



# jupyter-require
# Copyright 2019 Marek Cermak <macermak@redhat.com>
#
# MIT License
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Common notebook utilities."""

import csscompressor

from IPython import get_ipython

from .core import execute_with_requirements
from .core import safe_execute


Jupyter = get_ipython()
"""Current InteractiveShell instance."""


def link_css(href: str, attrs: dict = None):
    """Link CSS stylesheet."""
    script = """
        'use strict';
        
        const href = "$$href";
        const attributes = $$attrs || {};

        if ( !$(`link[href*="${href}"]`).length ) {
            let link = document.createElement("link");

            link.rel = "stylesheet";
            link.type = "text/css";
            try {
                link.href = requirejs.toUrl(href, 'css');
            } catch (error) {
                link.href = href;
            }
            
            Object.entries(attributes)
                .forEach( ([attr, val]) => $(link).attr(attr, val) );
            
            document.head.appendChild(link);
        }
    """

    return safe_execute(script, href=href, attrs=attrs)


def link_js(src: str):
    """Link JavaScript library."""
    script = """
        'use strict';

        const src = "$$src";

        if ( $("script").filter((i, e) => e.src.includes(src)).length <= 0 ) {
            let script = document.createElement("script");
            script.src = src;
            script.type = "text/javascript";

            document.head.appendChild(script);
        }
    """

    return safe_execute(script, src=src)


def load_css(style: str, attrs: dict = None, compress=True, **compressor_options):
    """Create new style element and add it to the page."""
    attrs = attrs or {}

    script = """
        'use strict'
        
        const style = `$$style`;
        const attributes = $$attrs || {};
        
        let id = attributes.id;
        let elem_exists = id ? $(`style#${id}`).length > 0 : false;
        
        let e = elem_exists ? document.querySelector(`style#${id}`)
                            : document.createElement(\"style\");
        
        $(e).text(`${style}`).attr('type', 'text/css');
        
        Object.entries(attributes)
            .forEach( ([attr, val]) => $(e).attr(attr, val) );

        if (!elem_exists) document.head.appendChild(e);
    """

    if compress:
        style = csscompressor.compress(style, **compressor_options)

    return safe_execute(script, style=style, attrs=attrs)


def load_js(script: str, attrs: dict = None):
    """Create new script element and add it to the page."""
    attrs = attrs or {}

    # escape dollar signs inside ticks and ticks
    js = script \
        .replace('`', '\`') \
        .replace('${', '\${')

    script = """
        'use strict';
    
        const script = `$$js`;
        const attributes = $$attrs || {};
        
        let id = attributes.id;
        let elem_exists = id ? $(`script#${id}`).length > 0 : false;
        
        let e = elem_exists ? document.querySelector(`script#${id}`)
                            : document.createElement(\"script\");
        
        $(e).text(`${script}`).attr('type', 'text/javascript');
        
        Object.entries(attributes)
            .forEach( ([attr, val]) => $(e).attr(attr, val) );

        if (!elem_exists) document.head.appendChild(e);
    """

    return safe_execute(script, js=js, attrs=attrs)


def enable_nbextension(nbextension):
    """Enable Jupyter nbextension."""
    script = """
    Jupyter.notebook.config.update({
        'load_extensions': {
            '$$nbextension': true
        }
    });
    """

    return safe_execute(script, nbextension=nbextension)


def disable_nbextension(nbextension):
    """Enable Jupyter nbextension."""
    script = """
    Jupyter.notebook.config.update({
        'load_extensions': {
            '$$nbextension': false
        }
    });
    """

    return safe_execute(script, nbextension=nbextension)


def load_nbextension(nbextension, endpoint: str = 'extension', enable=True):
    """Load Jupyter nbextension directly into the current notebook.

    Note: The extension has to be visible by notebook webserver, hence it has
    to be placed in defined folder locations (see the official [documentation])

    [documentation]: https://jupyter.readthedocs.io/en/latest/projects/jupyter-directories.html#jupyter-path
    """
    script = """
    utils.load_extensions('$$nbextension/$$endpoint');
    """

    if enable:
        enable_nbextension(nbextension)

    return execute_with_requirements(
        script, required=['base/js/utils'], nbextension=nbextension, endpoint=endpoint)


def clear_cell_metadata(index: int = None):
    """Clear cell requirement metadata."""
    script = """
    const index = $$index;
    
    let cells = Number.isInteger(index) ? [Jupyter.notebook.get_cell($$index)]
                                        : Jupyter.notebook.get_cells();
    
    cells.forEach((d, i) => {
        if (d.metadata.hasOwnProperty('require')) delete d.metadata.require;
    });
    """

    return safe_execute(script, index=index)


def clear_notebook_metadata():
    """Clear notebook requirement metadata."""
    script = """
    let nb = Jupyter.notebook;
    
    if (nb.metadata.hasOwnProperty('require')) delete nb.metadata.require;
    """

    return safe_execute(script)
